# DRAG_KNIFE.PY
# Rev: 15-July-2015
# Bill Trondsen
# billtrondsen@yahoo.com
###########################################################################
# This CAMBAM script creates a polyline for Drag-Knife cutting
# Select all desired polylines and run the script
# Note: Material Surface set to Z=0
# Use an engraving path with depth of cut set to zero
# Drag-Knife Path Concept comes from:  www.DonekTools.com
# Cut path is simulated by the path of a rear bicycle wheel
# Knowing the rear wheel path, calculate the front wheel (tool path)
# Formulas come from: http://math.ucsd.edu/~ebender/87/bicycle.pdf
#
# Release: 23-Sep-2013 Bill Trondsen
# Rev: 4-May-2014 Bill Trondsen
#      - fix Acos precision errors by adding a round() function
#      - add SwivelAngle for precision control
#      - add RetractEnable toggle to prepare for pluginizing
#      - resequence point logic to include RetractEnable and SwivelAngle
# Rev: 6-May-2014 Bill Trondsen
#      - fix error in FirstSegment with RetractEnable=0
#      - Point placed should be CFirst not C1
# Rev: 27-Jul-2014 Bill Trondsen
#      - Added DentLength variable and compensation, requested by Dragonfly
# Rev: 3-June-2015 by GraphicMan
#	   - Fixed the ramped swivel action (now swivel is done completely on the retract depth)
#	   - Log disabled for much faster calculations
#	   - Converts all polylines direction to CW
# Rev: 15-Jul-2015 Bill Trondsen
#      - Corrected VARIABLE description of CutDepth and RetractDepth to require NEGATIVE values
#
###########################################################################
#  VARIABLES:  Set these values for your tool and material
#
#    KnifeTipOffset: knife tip distance from tool centerline
#        (typically this is 1/2 the tool diameter)
#
#    CutDepth: set to material thickness or slightly larger to assure cutting through the material
#        Use a NEGATIVE value for CutDepth as the Z distance below the top surface of material.
#        This value would be equivalent to the Engrave Depth.
#
#    SwivelAngle: this angle is used as a precision factor
#        If the angle between segments is SMALLER than SwivelAngle, a Swivel move will NOT be produced.
#        This will make the cut smoother, but less precise to the original path.
#        Default value is 20 degrees.
#
#    DentLength: swivel paths are not placed if the segment is shorter than DentLength value
#        This feature will help to compensate for small jagged "Dents" on some scanned paths for instance
#        Requested by forum member Dragonfly
#        Similar in effect to builtin polyline function "Remove Overlaps"
#        Warning - could remove correct swivels if path contains short segments
#        Suggested default is 0
#
#    RetractEnable: this is a toggle to enable the Retract Feature
#        Value of (1) enables Retract and (0) disables it
#
#    RetractAngle: this angle is used for the Retract feature
#        If the angle between segments is LARGER than RetractAngle, a Retract move will be produced.
#        For thicker material like cardboard, this prevents gouging of material during the swivel move.
#        Default value is 40 degrees
#
#    RetractDepth: a smaller retract depth used during the swivel move
#        Use a NEGATIVE value for RetractDepth as the Z distance below the top surface of material.
#        As a suggestion, try 1/2 the value of the CutDepth.
#
###########################################################################


KnifeTipOffset = .25
CutDepth = 0
SwivelAngle = 20
DentLength = .02
RetractEnable = 0
RetractAngle = 40
RetractDepth = 0.05

# Convert All To CW################################

for ent in view.Selection:
  if ent.GetType().Name == "Polyline":
    # poly = Polyline(ent)
    poly = ent
    if poly.Direction.ToString() == "CCW":
      poly.Reverse()

###################################################

app.log("KnifeTipOffset "+KnifeTipOffset.ToString())
app.log("CutDepth       "+CutDepth.ToString())
app.log("SwivelAngle    "+SwivelAngle.ToString())
app.log("RetractEnable  "+RetractEnable.ToString())
app.log("RetractAngle   "+RetractAngle.ToString())
app.log("RetractDepth   "+RetractDepth.ToString())


####### MAIN LOOP #########################################

CamBamUI.MainUI.UndoBuffer.AddUndoPoint("Undo Here")
CamBamUI.MainUI.UndoBuffer.Add(doc.ActiveLayer.Entities)

for ent in view.Selection:
  if ent.GetType().Name == "Polyline":
    xpoly = XPolyline(ent)
    seg = xpoly.FirstSegment
    poly = Polyline()
    SegNum = 1

  ######## Loop for each segment within the Polyline #########
    while seg is not None:

      P1X = seg.P1.X
      P1Y = seg.P1.Y
      P2X = seg.P2.X
      P2Y = seg.P2.Y
      V1 = seg.VNormal1.Normal().Normal().Normal()
      V2 = seg.VNormal2.Normal().Normal().Normal()
      VFirst = Vector2F(1,0)
      VLast = Vector2F(1,0)

  #### First Point is oriented to X-axis and calculate sweep/bulge to first segment
      if seg == xpoly.FirstSegment:
        CFirst = Point2F(P1X+KnifeTipOffset, P1Y)
        CFirstXP = Vector3F.CrossProduct(Vector3F(1,0,0),Vector3F(V1.X,V1.Y,0))
        CFirstDirection = Math.Sign(CFirstXP.Z)
        if CFirstDirection == 0:
          CFirstDirection = 1
        CFirstSweep = Math.Acos(round(Vector2F.DotProduct(VFirst,V1)/(VFirst.Length*V1.Length),10))
        CFirstB = Math.Tan(CFirstSweep/4)*CFirstDirection

  #### Set the VN1 vector direction of the next segment

      if seg == xpoly.LastSegment:  
        VN1 = VLast    # vector VLast points toward +X after last segment

      else:
        VN1 = seg.NextSegment.VNormal1.Normal()
        VN1.Invert()


  #### Remaining Segments - Calculate Cut Points (C1,C2) and Cut Segment Bulge (C1B)

      C1 = Point2F(P1X+V1.X*KnifeTipOffset,P1Y+V1.Y*KnifeTipOffset)

      if seg.GetType().Name == "XPolySegLine":      
        C1B = 0
        C2 = Point2F(P2X+V2.X*KnifeTipOffset, P2Y+V2.Y*KnifeTipOffset)

      elif seg.GetType().Name == "XPolySegArc":
        C1B = Math.Tan(seg.Sweep/4)
        C2X = P2X+KnifeTipOffset/((1+(V2.Y/V2.X)**2)**.5)*Math.Sign(V2.X)
        C2Y = P2Y+(KnifeTipOffset*V2.Y/V2.X)/((1+(V2.Y/V2.X)**2)**.5)*Math.Sign(V2.X)
        C2 = Point2F(C2X,C2Y)


  #### Calculate Bulge (C2B) for the Sweep Path between segments

      VC2 = Vector2F(seg.P2,C2)  # vector from P2 to C2 used for calculating sweep path bulge
      C2XP = Vector3F.CrossProduct(Vector3F(VC2.X,VC2.Y,0),Vector3F(VN1.X,VN1.Y,0))
      C2Direction = Math.Sign(C2XP.Z)
      if C2Direction == 0:
        C2Direction = 1
      C2Sweep = Math.Acos(round(Vector2F.DotProduct(VC2,VN1)/(VC2.Length*VN1.Length),10))
      C2B = Math.Tan((C2Sweep)/4)*C2Direction
      C2Angle = C2Sweep*180/Math.PI


  #### Calculate CLast to orient the tool in the +X direction to prepare for additional segments

      CLast = Point2F(seg.P2.X+VLast.X*KnifeTipOffset, seg.P2.Y+VLast.Y*KnifeTipOffset)


  ######  End of Calculation Section


########################################################
# Add points to path with depth logic###################
########################################################

      if seg == xpoly.FirstSegment:
        if RetractEnable == 1:
          poly.Add(CFirst.X, CFirst.Y, RetractDepth, CFirstB)
          poly.Add(C1.X, C1.Y, RetractDepth, 0)
        elif RetractEnable == 0:
          poly.Add(CFirst.X, CFirst.Y, CutDepth, CFirstB)

      if seg.Length > DentLength: # Ignore placement of short "Dent" segments
          poly.Add(C1.X, C1.Y, CutDepth, C1B)

      if C2Angle >= SwivelAngle:  # then do a swivel move
        if seg.Length > DentLength: # Ignore swivel segment for "Dents"
          if RetractEnable == 1:
            poly.Add(C2.X, C2.Y, CutDepth, 0)
            poly.Add(C2.X, C2.Y, RetractDepth, C2B)
            poly.Add(P2X+KnifeTipOffset*VN1.X, P2Y+KnifeTipOffset*VN1.Y, RetractDepth, 0)
          elif RetractEnable == 0:
            poly.Add(C2.X, C2.Y, CutDepth, C2B)

      else: # otherwise do not do a swivel move
        poly.Add(P2X+KnifeTipOffset*VN1.X, P2Y+KnifeTipOffset*VN1.Y, CutDepth, 0)

      if seg == xpoly.LastSegment:
        if RetractEnable == 1:
          poly.Add(CLast.X, CLast.Y, RetractDepth, 0)
        elif RetractEnable == 0:
          poly.Add(CLast.X, CLast.Y, CutDepth, 0)

#### End of Main Loop #####################################

      SegNum += 1
      seg = seg.NextSegment
    doc.Add(poly)